# pylint: disable=unused-import
import logging

from django.apps import AppConfig

logger = logging.getLogger(__name__)


class SideEffectsConfig(AppConfig):

    name = "side_effects"
    verbose_name = "External Side Effects"
    configs = []

    def ready(self):
        logger.debug("Initialising side_effects registry")
        from . import registry

        logger.debug("Registering side_effects checks")
        from . import checks



from django.conf import settings
from env_utils import get_bool


def get_setting(setting_name, default_value):
    return getattr(settings, setting_name, get_bool(setting_name, default_value))


# If True then instead of logging exceptions in side-effects
# functions, raise them, which will abort processing.
# Default = False
ABORT_ON_ERROR = get_setting("SIDE_EFFECTS_ABORT_ON_ERROR", False)

# In test mode no side-effects are run
# Default = False
TEST_MODE = get_setting("SIDE_EFFECTS_TEST_MODE", False)

# In FAIL test mode any call to run_side_effects will raise an Exception
# This is used to uncover any tests that are running side-effects when
# they shouldn't be.
# Default = False
TEST_MODE_FAIL = get_setting("SIDE_EFFECTS_TEST_MODE_FAIL", False)



import json
from os import sys

from django.core.management.base import BaseCommand
from side_effects.registry import _registry, docstring, fname


class Command(BaseCommand):

    help = "Displays project side_effects."

    def __init__(self, *args, **kwargs):
        self.missing_docstrings = []
        super().__init__(*args, **kwargs)

    def add_arguments(self, parser):
        parser.add_argument(
            "--raw",
            action="store_true",
            help="Display raw mapping of labels to functions.",
        )
        parser.add_argument(
            "--verbose",
            action="store_true",
            help="Display full docstring for all side-effect functions.",
        )
        parser.add_argument(
            "--strict",
            action="store_true",
            default=False,
            dest="strict",
            help="Exit with a non-zero exit code if any registered functions have no docstrings.",
        )
        parser.add_argument(
            "--label",
            action="store",
            dest="label",
            help="Filter side-effects on a single event label.",
        )
        parser.add_argument(
            "--label-contains",
            action="store",
            dest="label-contains",
            help="Filter side-effects on event labels containing the supplied value.",
        )

    def handle(self, *args, **options):
        if options["label"]:
            self.stdout.write(
                f"\nSide-effects for event matching '{options['label']}':"
            )
            events = _registry.by_label(options["label"]).items()
        elif options["label-contains"]:
            self.stdout.write(
                f"\nSide-effects for events matching '*{options['label-contains']}*':"
            )
            events = _registry.by_label_contains(options["label-contains"]).items()
        else:
            self.stdout.write("\nRegistered side-effects:")
            events = _registry.items()

        if options["raw"]:
            self.print_raw(events)
        elif options["verbose"]:
            self.print_verbose(events)
        else:
            self.print_default(events)

        self.print_missing()

        if options["strict"]:
            self.exit()

    def print_raw(self, events: dict) -> None:
        """Print out the fully-qualified named for each mapped function."""
        raw = {label: [fname(f) for f in funcs] for label, funcs in events}
        self.stdout.write(json.dumps(raw, indent=4))

    def print_verbose(self, events: dict) -> None:
        """Print the entire docstring for each mapped function."""
        for label, funcs in events:
            self.stdout.write("")
            self.stdout.write(label)
            self.stdout.write("")
            for func in funcs:
                docs = docstring(func)
                if docs is None:
                    self.missing_docstrings.append(fname(func))
                    self.stderr.write(f"  x {fname(func)} (no docstring)")
                    self.stdout.write("")
                else:
                    self.stdout.write(f"  - {fname(func)}:")
                    self.stdout.write(f"    {docs[0]}")
                    for line in docs[1:]:
                        self.stdout.write(f"    {line}")
                    self.stdout.write("")

    def print_default(self, events: dict) -> None:
        """Print the first line of the docstring for each mapped function."""
        for label, funcs in events:
            self.stdout.write("")
            self.stdout.write(label)
            for func in funcs:
                docs = docstring(func)
                if docs is None:
                    self.missing_docstrings.append(fname(func))
                    self.stderr.write(f"  x {fname(func)} (no docstring)")
                else:
                    self.stdout.write(f"  - {docs[0]}")

    def print_missing(self):
        """Print out the contents of self.missing_docstrings."""
        if self.missing_docstrings:
            self.stderr.write("\nThe following functions have no docstrings:")
            for md in self.missing_docstrings:
                self.stderr.write(f"  {md}")
        else:
            self.stdout.write("\nAll registered functions have docstrings")

    def exit(self):
        """Exit based on whether there are any missing docstrings.

        This is used in CI scenarios to fail a build in which there are missing
        docstrings.

        """
        sys.exit(len(self.missing_docstrings))



from functools import wraps

from django.http import HttpResponse

from . import registry


def http_response_check(response):
    """Return True for anything other than 4xx, 5xx status_codes."""
    if isinstance(response, HttpResponse):
        return not (400 <= response.status_code < 600)
    else:
        return True


def has_side_effects(label: str, run_on_exit: bool = http_response_check):
    """
    Run decorated function and raise side_effects signal when complete.

    This decorator should be used to indicate that a function has external
    side-effects that should be processed after the function has returned -
    you can think of it as a universal post_FOO decorator for any function.

    In its current form it calls the signal handler inside of any external
    transaction - so any error raised will abort the transaction - so please
    bear this in mind.

    Any function decorated with this will be picked up in the side_effects
    module in core. You can use the label to call the appropriate side
    effect function.

    The run_on_exit kwarg can be used for fine-grained control over the exact
    behaviour required. The canonical use case for this is when decorating
    view functions - as they will typically always return a valid HttpResponse
    object, and use the status_code property to indicate whether the view
    function ran OK.

    Args:
        label: string, an identifier that is used in the receiver to determine
            which event has occurred. This is required because the function name
            won't be sufficient in most cases.

    Kwargs:
        run_on_exit: function used to determine whether the side effects should
            run, based on the return value of the innner function. This can be
            used to inspect the result for fine grained control. The default is
            `http_response_check`, which will return False for 4xx, 5xx status
            codes.

    """

    def decorator(func):
        @wraps(func)
        def inner_func(*args, **kwargs):
            """Run the original function and send the signal if successful."""
            return_value = func(*args, **kwargs)
            if run_on_exit(return_value):
                registry.run_side_effects(
                    label, *args, return_value=return_value, **kwargs
                )
            return return_value

        return inner_func

    return decorator


def is_side_effect_of(label):
    """Register a function as a side-effect."""

    def decorator(func):
        registry.register_side_effect(label, func)

        @wraps(func)
        def inner_func(*args, **kwargs):
            return func(*args, **kwargs)

        return inner_func

    return decorator


def disable_side_effects():
    """Disable side-effects from firing - used for testing."""

    def decorator(func):
        @wraps(func)
        def inner_func(*args, **kwargs):
            with registry.disable_side_effects() as events:
                return func(*args, events, **kwargs)

        return inner_func

    return decorator



"""
This module contains the Registry class that is responsible for managing
all of the registered side-effects.
"""
import inspect
import logging
import threading
from collections import defaultdict

from django.dispatch import Signal

from . import settings

logger = logging.getLogger(__name__)


def fname(func):
    """Return fully-qualified function name."""
    return "%s.%s" % (func.__module__, func.__name__)


def docstring(func):
    """Split and strip function docstrings into a list of lines."""
    try:
        lines = func.__doc__.strip().split("\n")
        return [line.strip() for line in lines]
    except AttributeError:
        return None


class SideEffectsTestFailure(Exception):
    def __init__(self, label):
        super().__init__(f"Side-effects for '{label}' aborted; TEST_MODE_FAIL=True")


class SignatureMismatch(Exception):
    def __init__(self, func):
        super().__init__(
            f"Side-effect signature mismatch for function `{func.__name__}{inspect.signature(func)}`."
        )

    pass


class Registry(defaultdict):

    """
    Registry of side effect functions.

    This class is a defaultdict(list) that contains a
    mapping of the side-effect label to the functions
    that should run after the function has completed.

    It has two additional methods - `add`, which is used
    to register a new function against a label, and
    `contains` which is used to look up a function against
    a label.

    """

    # if using the disable_side_effects context manager or decorator,
    # then this signal is used to communicate details of events that
    # would have fired, but have been suppressed.
    suppressed_side_effect = Signal(providing_args=["label"])

    def __init__(self):
        self._lock = threading.Lock()
        self._suppress = False
        super(Registry, self).__init__(list)

    def by_label(self, value):
        """Filter registry by label (exact match)."""
        return {k: v for k, v in self.items() if k == value}

    def by_label_contains(self, value):
        """Filter registry by label (contains string)."""
        return {k: v for k, v in self.items() if value in k}

    def contains(self, label, func):
        """
        Lookup label: function mapping in the registry.

        The fname of the function is used in the lookup, as running
        a simple `func in list` check doesn't work.

        """
        return fname(func) in [fname(f) for f in self[label]]

    def add(self, label, func):
        """
        Add a function to the registry.

        Args:
            label: string, the name of the side-effect - this is used
                to bind two function - the function that @has_side_effects
                and the function that @is_side_effect.
            func: a function that will be called when the side-effects are
                executed - this will be passed all of the args and kwargs
                of the original function.

        """
        with self._lock:
            self[label].append(func)

    def _run_side_effects(self, label, *args, return_value=None, **kwargs):
        if settings.TEST_MODE_FAIL:
            raise SideEffectsTestFailure(label)
        for func in self[label]:
            _run_func(func, *args, return_value=return_value, **kwargs)

    def run_side_effects(self, label, *args, return_value=None, **kwargs):
        """Run registered side-effects functions, or suppress as appropriate.

        If TEST_MODE is on, or the _suppress attr is True, then the side-effects
        are not run, but the `suppressed_side_effect` signal is sent - this is
        primarily used by the disable_side_effects context manager to register
        which side-effects events were suppressed (for testing purposes).

        NB even if the side-effects themselves are not run, this method will try
        to bind all of the receiver functions - this is to ensure that incompatible
        functions fail hard and early.

        """
        # TODO: this is all becoming over-complex - need to simplify this
        self.try_bind_all(label, *args, return_value=return_value, **kwargs)
        if self._suppress or settings.TEST_MODE:
            self.suppressed_side_effect.send(Registry, label=label)
        else:
            self._run_side_effects(label, *args, return_value=return_value, **kwargs)

    def try_bind_all(self, label, *args, return_value=None, **kwargs):
        """
        Test all receivers for signature compatibility.

        Raise SignatureMismatch if any function does not match.

        """
        for func in self[label]:
            if not (
                try_bind(func, *args, return_value=return_value, **kwargs)
                or try_bind(func, *args, **kwargs)
            ):
                raise SignatureMismatch(func)


class disable_side_effects:

    """Context manager used to disable side-effects temporarily.

    This works by setting the _suppress attribute on the registry object,
    and then connecting a receiver to the Signal that emits details of
    events that were suppressed.

    NB this changes global state and, ironically, may have unintended consequences

    """

    def __init__(self):
        self.events = []
        pass

    def __enter__(self):
        _registry.suppressed_side_effect.connect(self.on_event, dispatch_uid="suppress")
        _registry._suppress = True
        return self.events

    def __exit__(self, *args):
        _registry._suppress = False
        _registry.suppressed_side_effect.disconnect(self.on_event)

    def on_event(self, sender, **kwargs):
        self.events.append(kwargs["label"])


def register_side_effect(label, func):
    """Helper function to add a side-effect function to the registry."""
    if func in _registry[label]:
        return
    _registry.add(label, func)


def run_side_effects(label, *args, return_value=None, **kwargs):
    """Run all of the side-effect functions registered for a label."""
    _registry.run_side_effects(label, *args, return_value=return_value, **kwargs)


def _run_func(func, *args, return_value=None, **kwargs):
    """Run a single side-effect function and handle errors."""
    try:
        if try_bind(func, *args, return_value=return_value, **kwargs):
            func(*args, return_value=return_value, **kwargs)
        elif try_bind(func, *args, **kwargs):
            func(*args, **kwargs)
        else:
            raise SignatureMismatch(func)
    except SignatureMismatch:
        # always re-raise SignatureMismatch as this means we have been unable
        # to run the side-effect function at all.
        raise
    except Exception:
        logger.exception("Error running side_effect function '%s'", fname(func))
        if settings.ABORT_ON_ERROR or settings.TEST_MODE_FAIL:
            raise


def try_bind(func, *args, **kwargs):
    """Try binding args & kwargs to a given func."""
    try:
        inspect.signature(func).bind(*args, **kwargs)
    except TypeError:
        return False
    else:
        return True


# global registry
_registry = Registry()



from os import path, pardir, chdir
from setuptools import setup, find_packages

from side_effects import __version__

README = open(path.join(path.dirname(__file__), "README.rst")).read()
# allow setup.py to be run from any path
chdir(path.normpath(path.join(path.abspath(__file__), pardir)))

setup(
    name="django-side-effects",
    version=__version__,
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    description="Django app for managing external side effects.",
    long_description=README,
    url="https://github.com/yunojuno/django-side-effects",
    install_requires=["django>=1.11,<3.0", "python-env-utils"],
    author="YunoJuno",
    author_email="code@yunojuno.com",
    license="MIT",
    maintainer="YunoJuno",
    maintainer_email="code@yunojuno.com",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



import inspect

from django.core.checks import messages, register

from . import registry

REGISTRY = registry._registry
CHECK_ID_MULTIPLE_SIGNATURES = "side_effects.W001"


def _message(label: str) -> messages.CheckMessage:
    """Create Error or Warning message based on STRICT_MODE."""
    msg = f'Multiple function signatures for event: "{label}"'
    hint = (
        f"Ensure that all functions decorated "
        f'`@is_side_effect_of("{label}")` have identical signatures.'
    )
    return messages.Warning(msg, hint=hint, id=CHECK_ID_MULTIPLE_SIGNATURES)


def signature_count(label: str) -> int:
    """Number of unique function signatures for an event."""
    signatures = [inspect.signature(func) for func in registry._registry[label]]
    return len(set(signatures))


@register()
def check_function_signatures(app_configs, **kwargs):
    """Check that all registered functions have the same signature."""
    errors = []
    for label in REGISTRY:
        if signature_count(label) > 1:
            errors.append(_message(label))
    return errors
